import peewee


__all__ = ('Handle',)


class Handle:

    __slots__ = ('_seed', '_authorize', '_error', '_parse')

    def __init__(self, seed, authorize, error, parse):

        self._seed = seed

        self._authorize = authorize

        self._error = error

        self._parse = parse

    @property
    def seed(self):

        return self._seed

    async def _execute(self, action, request, accept):

        self._authorize(request)

        try:

            data = await request.json()

        except:

            raise self._error('invalid data')

        method = request.method

        path = request.url.path

        (keys, *rest) = await self._parse(method, data)

        try:

            data = await action(keys, *rest)

        except peewee.DatabaseError as error:

            raise self._error('database fail', str(error))

        return data

    def select(self, request):

        accept = (list, list) # [keys, names]

        return self._execute(self._seed.select, request, accept)

    def create(self, request):

        accept = (list, dict) # [keys, data]

        return self._execute(self._seed.create, request, accept)

    def update(self, request):

        accept = (list, dict) # [keys, data]

        return self._execute(self._seed.update, request, accept)

    def delete(self, request):

        accept = (list,) # [keys]

        return self._execute(self._seed.delete, request, accept)



import asyncio
import aiohttp.web
import collections
import functools

from . import state


__all__ = ('Client',)


class Client:

    __slots__ = ('_authorize', '_infos', '_interval', '_states', '_loop')

    __marker = object()

    def __init__(self, authorize, interval, loop = None):

        if not loop:

            loop = asyncio.get_event_loop()

        self._authorize = authorize

        self._infos = []

        self._interval = interval

        self._states = []

        self._loop = loop

    def inform(self, name, keys):

        info = (name, keys)

        self._infos.append(info)

    async def dispatch(self, action, name, payload):

        data = (action, name, payload)

        coroutines = (state.dispatch(data) for state in self._states)

        await asyncio.gather(*coroutines, loop = self._loop)

    async def connect(self, request):

        self._authorize(request)

        websocket = aiohttp.web.WebSocketResponse()

        await websocket.prepare(request)

        value = state.State(websocket, loop = self._loop)

        self._states.append(value)

        await value.hello(self._interval, self._infos)

        try:

            await value.start()

        finally:

            self._states.remove(value)

        return websocket

    async def close(self):

        for state in self._states:

            await state.close()





__all__ = ('RequestError', 'AuthorizationError', 'BadRequestError')


class RequestError(Exception):

    __slots__ = ('response', 'code', 'info')

    def __init__(self, response, code, info):

        self.response = response

        self.code = code

        self.info = info

    def __repr__(self):

        return f'{self.__class__.__name__}: {self.code} {self.info}'


class AuthorizationError(RequestError):

    __slots__ = ()


class BadRequestError(RequestError):

    __slots__ = ()



import asyncio
import aiohttp.web
import functools

from . import access
from . import seed
from . import handle


__all__ = ()


def authorization(error, key, value, store):

    try:

        provide = store[key]

    except KeyError:

        raise error('missing')

    if provide == value:

        return

    raise error('invalid')


def fail(cls, *data):

    response = aiohttp.web.json_response(data)

    return cls(body = response.body, content_type = response.content_type)


authorization_header = 'Authorization'


authorization_error = aiohttp.web.HTTPUnauthorized


authorization_fail = functools.partial(fail, authorization_error)


authorization_parts = (authorization_fail, authorization_header)


specific_authorize = functools.partial(authorization, *authorization_parts)


def specific_authorization(spot, value, request):

    store = getattr(request, spot)

    specific_authorize(value, store)


authorization_spot = 'headers'


authorize = functools.partial(specific_authorization, authorization_spot)


authorizer = lambda value: functools.partial(authorize, value)


handle_error = aiohttp.web.HTTPBadRequest


handle_fail = functools.partial(fail, authorization_error)


ignore = ('GET',)


methods = (*ignore, 'POST', 'PATCH', 'DELETE')


schemas = ((list, list), (list, dict), (list, dict), (list,))


accept = dict(zip(methods, schemas))


async def parse(error, method, data):

    schema = accept[method]

    if not len(data) == len(schema):

        raise error('invalid data length')

    values = []

    for (data, cls) in zip(data, schema):

        if not isinstance(data, cls):

            raise error('invalid data type')

        values.append(data)

    return values


def create(client, objects, model, authorize, fail, parse = parse):

    access_ = access.Access(objects, model)

    client.inform(access_.name, access_.primary)

    seed_ = seed.Seed(access_, tuple)

    def _parse(method, data):

        return parse(fail, method, data)

    handle_ = handle.Handle(seed_, authorize, fail, _parse)

    return handle_


def listen(action, alert):

    async def wrapper(request):

        data = await action(request)

        if alert:

            await alert(data)

        return aiohttp.web.json_response(data)

    return wrapper


def funnel(client, handle):

    name = handle.seed.access.name

    actions = (handle.select, handle.create, handle.update, handle.delete)

    path = '/' + name

    def vocal(*args):

        def wrapper(data):

            return client.dispatch(*args, data)

        return wrapper

    for (method, action) in zip(methods, actions):

        dispatch = None if method in ignore else vocal(method, name)

        action = listen(action, dispatch)

        yield (method, path, action)


def setup(client, *args, **kwargs):

    handle = create(client, *args, **kwargs)

    yield from funnel(client, handle)



import functools


__all__ = ()


def authorization(spot, key, value, options):

    store = options.setdefault(spot, {})

    store[key] = value


authorization_spot = 'headers'


authorization_header = 'Authorization'


authorization_parts = (authorization_spot, authorization_header)


authorize = functools.partial(authorization, *authorization_parts)


authorizer = lambda value: functools.partial(authorize, value)



import itertools

from . import helpers


__all__ = ()


class Entry:

    __slots__ = ('__dict__',)

    __init__ = helpers.update_value

    def __getitem__(self, key):

        try:

            value = getattr(self, key)

        except AttributeError:

            raise KeyError(key) from None

        return value


class Entries(dict):

    __slots__ = ('_primary',)

    def __init__(self, primary):

        self._primary = primary

    @property
    def primary(self):

        return self._primary

    def get(self, *keys):

        value = self

        try:

            (*values, value) = helpers.crawl(value, keys)

        except ValueError:

            pass

        return value

    def query(self, data):

        return (data[key] for key in self._primary)

    def create(self, keys, data):

        (*start, final) = keys

        value = self

        for (index, key) in enumerate(start):

            try:

                value = value[key]

            except KeyError:

                pass

            else:

                continue

            value[key] = value = self.__class__(self._primary[index + 1:])

        value[final] = entry = Entry(data)

        return entry

    def destroy(self, keys):

        path, keys = itertools.tee(keys)

        values = tuple(helpers.crawl(self, keys))

        origins, values = itertools.tee(values)

        stores = (self, *origins)

        bundles = zip(stores, path, values)

        orderly = reversed(tuple(bundles))

        for (index, bundle) in enumerate(orderly):

            (store, key, value) = bundle

            if index:

                if value:

                    break

            else:

                entry = value

            del store[key]

        return entry




from .access import *
from .seed import *
from .handle import *
from .client import *

from . import utils


__all__ = (*access.__all__, *seed.__all__, *handle.__all__, *client.__all__,
           'utils')





__all__ = ('Seed',)


class Seed:

    __slots__ = ('_access', '_parse')

    def __init__(self, access, parse):

        self._access = access

        self._parse = parse

    @property
    def access(self):

        return self._access

    async def select(self, keys, names):

        data = await self._access.select(keys, names)

        return self._parse(data)

    async def create(self, keys, data):

        await self._access.create(keys, data)

        data = await self._access.select(keys, ())

        return self._parse(data)

    async def update(self, keys, data):

        await self._access.update(keys, data)

        names = data.keys()

        data = await self._access.select(keys, ())

        return self._parse(data)

    async def delete(self, keys):

        data = await self._access.select(keys, ())

        await self._access.delete(keys)

        return self._parse(data)





__all__ = ()


def update_value(value, data):

    existing = set(value.__dict__.keys())

    incoming = set(data.keys())

    outgoing = existing - incoming

    for key in outgoing:

        delattr(value, key)

    for key in incoming:

        setattr(value, key, data[key])


def crawl(value, keys):

    for key in keys:

        value = value[key]

        yield value



import asyncio
import aiohttp


__all__ = ('State',)


class Heart:

    __slots__ = ('_ack', '_kill', '_beated', '_closed', '_task', '_loop')

    def __init__(self, ack, kill, loop):

        self._ack = ack

        self._kill = kill

        self._beated = None

        self._closed = None

        self._task = None

        self._loop = loop

    async def ack(self):

        self._beated = True

        await self._ack()

    async def pulse(self, interval):

        while not self._closed:

            self._beated = False

            await asyncio.sleep(interval)

            if self._beated:

                continue

            await self._kill()

            self.close()

    def start(self, interval):

        if not self._task is None:

            self._task.cancel()

        self._closed = False

        coroutine = self.pulse(interval)

        self._task = self._loop.create_task(coroutine)

    def close(self):

        self._closed = True

        if self._task is None:

            return

        self._task.cancel()


class Codes:

    dispatch = 0
    hello = 1
    beat = 2
    ack = 3


class State:

    __slots__ = ('_websocket', '_heart')

    def __init__(self, websocket, loop = None):

        if not loop:

            loop = asyncio.get_event_loop()

        self._websocket = websocket

        self._heart = Heart(self._ack, self.close, loop)

    async def _push(self, code, payload):

        data = (code, payload)

        await self._websocket.send_json(data)

    async def _ack(self):

        await self._push(Codes.ack, None)

    async def hello(self, interval, information):

        data = (interval, information)

        pulsing = self._heart.start(interval)

        await self._push(Codes.hello, data)

    async def dispatch(self, data):

        await self._push(Codes.dispatch, data)

    async def _pull(self, data):

        (code, payload) = data

        if code == Codes.beat:

            await self._heart.ack()

            return

    async def stream(self):

        while not self._websocket.closed:

            try:

                data = await self._websocket.receive_json()

            except (asyncio.CancelledError, TypeError):

                break

            try:

                await self._pull(data)

            except:

                continue

    async def close(self):

        await self._websocket.close()

        self._heart.close()

    async def start(self):

        try:

            await self.stream()

        finally:

            await self.close()



import asyncio
import aiohttp
import yarl
import collections
import functools
import copy
import aiocogs

from . import errors
from . import state
from . import cache
from . import helpers


__all__ = ('Client',)


class Client:

    __slots__ = ('_session', '_url', '_authorize')

    _errors = {
        401: errors.AuthorizationError,
        400: errors.BadRequestError
    }

    _valid = (*_errors.keys(), 200)

    def __init__(self, session, url, authorize):

        self._session = session

        self._url = yarl.URL(url)

        self._authorize = authorize

    async def request(self, method, path, **kwargs):

        self._authorize(kwargs)

        clean = self._url.path.rstrip('/')

        url = self._url.with_path(clean + path)

        response = await self._session.request(method, url, **kwargs)

        data = await response.json() if response.status in self._valid else None

        if response.status < 400:

            return data

        try:

            error = self._errors[response.status]

        except KeyError:

            error = errors.RequestError

        if data:

            code, *info = data

        else:

            code, info = (None,) * 2

        raise error(response, code, info)


class Client(Client):

    __slots__ = ('_cache', '_ignore', '_tracks', '_state', '_ready', '_closed',
                 '_loop')

    def __init__(self, *args, ignore = ().__contains__, loop = None):

        super().__init__(*args)

        if not loop:

            loop = asyncio.get_event_loop()

        self._cache = None

        self._ignore = ignore

        self._tracks = collections.defaultdict(
            functools.partial(
                collections.defaultdict,
                list
            )
        )

        self._state = None

        self._ready = asyncio.Event(loop = loop)

        self._closed = False

        self._loop = loop

    @property
    def cache(self):

        return self._cache

    async def request(self, method, name, *data):

        path = '/' + name

        data = await super().request(method, path, json = data)

        return tuple(map(cache.Entry, data))

    def get(self, name, *keys, names = ()):

        return self.request('GET', name, keys, names)

    def create(self, name, *keys, **data):

        return self.request('POST', name, keys, data)

    def update(self, name, *keys, **data):

        return self.request('PATCH', name, keys, data)

    def delete(self, name, *keys):

        return self.request('DELETE', name, keys)

    _methods = {
        'create': 'POST',
        'update': 'PATCH',
        'delete': 'DELETE'
    }

    def track(self, action, name):

        method = self._methods[action]

        def wrapper(function):

            self._tracks[method][name].append(function)

            return function

        return wrapper

    def check(self, action, name):

        method = self._methods[action]

        value = asyncio.Event(loop = self._loop)

        def decorator(function):

            callbacks = self._tracks[method][name]

            async def callback(*args):

                if not await function(*args):

                    return

                value.set()

            callbacks.append(callback)

            async def observe():

                await value.wait()

                callbacks.remove(callback)

            coroutine = observe()

            self._loop.create_task(coroutine)

            return value

        return decorator

    def _ePOST(self, name, data):

        entries = self._cache[name]

        query = entries.query(data)

        entry = entries.create(query, data)

        return entry

    def _ePATCH(self, name, data):

        entries = self._cache[name]

        query = entries.query(data)

        entry = entries.get(*query)

        fake = copy.copy(entry)

        helpers.update_value(entry, data)

        return (fake, entry)

    def _eDELETE(self, name, data):

        entries = self._cache[name]

        query = entries.query(data)

        entry = entries.destroy(query)

        return entry

    async def handle(self, method, name, data, prefix = '_e'):

        await self._ready.wait()

        if self._ignore(name):

            entries = map(cache.Entry, data)

        else:

            handle = getattr(self, prefix + method)

            entries = (handle(name, value) for value in data)

        entries = tuple(entries)

        callbacks = self._tracks[method][name]

        coroutines = (callback(entries) for callback in callbacks)

        await asyncio.gather(*coroutines, loop = self._loop)

    async def _build(self, names):

        names = tuple(names)

        coroutines = map(self.get, names)

        tasks = tuple(map(self._loop.create_task, coroutines))

        async for task in aiocogs.ready(*tasks, loop = self._loop):

            values = task.result()

            index = tasks.index(task)

            name = names[index]

            store = self._cache[name]

            for value in values:

                data = value.__dict__

                query = store.query(data)

                store.create(query, data)

        self._ready.set()

    def _setup(self, data):

        store = {}

        for (name, key) in data:

            if self._ignore(name):

                continue

            store[name] = cache.Entries(key)

        self._cache = cache.Entry(store)

        return self._build(store.keys())

    def _callback(self, data, prefix = '_e'):

        if self._cache:

            (method, name, payload) = data

            coroutine = self.handle(method, name, payload)

        else:

            coroutine = self._setup(data)

        self._loop.create_task(coroutine)

    async def connect(self, path):

        (params, headers) = ({}, {})

        options = {'params': params, 'headers': headers}

        self._authorize(options)

        path = self._url.path.rstrip('/') + path

        url = self._url.with_path(path).with_query(params)

        websocket = await self._session.ws_connect(url, headers = headers)

        return state.State(websocket, self._callback, loop = self._loop)

    async def manage(self, execute, retry):

        while not self._closed:

            while not self._closed:

                try:

                    state = await execute()

                except aiohttp.ClientConnectionError:

                    await asyncio.sleep(retry)

                else:

                    break

            else:

                break

            self._state = state

            await self._state.start()

    async def start(self, path, retry = 0.5):

        connect = functools.partial(self.connect, path)

        coroutine = self.manage(connect, retry)

        self._loop.create_task(coroutine)

        await self._ready.wait()

    async def close(self):

        self._closed = True

        await self._state.close()



import operator
import functools


__all__ = ('Access',)


_containers = (tuple, list, set, frozenset)


class Access:

    __slots__ = ('_objects', '_model')

    def __init__(self, objects, model):

        self._objects = objects

        self._model = model

    @property
    def name(self):

        return self._model._meta.name

    @property
    def primary(self):

        keys = self._model._meta.get_primary_keys()

        return tuple(key.column_name for key in keys)

    @staticmethod
    def _compatible(keys):

        (*leads, finals) = keys

        if isinstance(finals, _containers):

            finals = tuple(finals)

        else:

            finals = (finals,)

        return (leads, finals)

    def query(self, source, keys):

        (leads, finals) = self._compatible(keys)

        fields = (getattr(self._model, key) for key in self.primary)

        if leads:

            associations = (field == key for (field, key) in zip(leads, fields))

            action = functools.reduce(operator.and_, associations)

            source = source.where(action)

        try:

            field = next(fields)

        except StopIteration:

            pass

        else:

            source = source.where(field << finals)

        return source

    def _revert(self, data):

        for (key, value) in data.items():

            if not value is None:

                continue

            data[key] = getattr(self._model, key).default

    def select(self, keys, names):

        columns = (getattr(self._model, name) for name in names)

        action = self._model.select(*columns).dicts()

        if keys:

            action = self.query(action, keys)

        return self._objects.execute(action)

    async def create(self, keys, data):

        self._revert(data)

        (leads, finals) = self._compatible(keys)

        series = ((*leads, final) for final in finals)

        extras = (zip(self.primary, keys) for keys in series)

        async with self._objects.atomic():

            for extra in extras:

                data.update(extra)

                action = self._model.insert(**data)

                await self._objects.execute(action)

    async def update(self, keys, data):

        self._revert(data)

        action = self._model.update(**data)

        if keys:

            action = self.query(action, keys)

        await self._objects.execute(action)

    async def delete(self, keys):

        action = self._model.delete()

        if keys:

            action = self.query(action, keys)

        await self._objects.execute(action)



import setuptools

with open('README.md') as file:

    readme = file.read()

name = 'aiodata'

version = '1.4.10'

author = 'Exahilosys'

url = f'https://github.com/{author}/{name}'

download_url = f'{url}/archive/v{version}.tar.gz'

setuptools.setup(
    name = name,
    version = version,
    author = author,
    author_email = 'exahilosys@gmail.com',
    url = url,
    download_url = download_url,
    packages = setuptools.find_packages(),
    license = 'MIT',
    description = 'HTTP database access.',
    long_description = readme,
    long_description_content_type = 'text/markdown',
    include_package_data = True,
    install_requires = [
        'aiohttp',
        'yarl',
        'aiocogs',
        'peewee'
    ],
    classifiers = [
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ]
)
